/*
 * Pam Coda Module
 * ALPHA - V0.2 
 * Robin.Gareus@urz.uni-heidelberg.de
 * http://www.rzuser.uni-heidelberg.de/~x42/coda/index.html
 *
 * 
 * This Software is distributed under the terms of the GNU 
 * General Public License (GPL)
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <grp.h>
#include <pwd.h>
#include <time.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/wait.h>
#include <sys/types.h>


#define PAM_SM_AUTH
#define CLOG_PROGRAM "/usr/local/bin/clog"

#include <security/pam_modules.h>
#include <security/_pam_macros.h>
extern int __get_authtok(pam_handle_t * pamh, int not_set_pass);
/* --- authentication management functions --- */

PAM_EXTERN
int pam_sm_authenticate(pam_handle_t * pamh, int flags, int argc
			,const char **argv)
{
    int retval;
    const char *user = NULL;
    struct passwd *pw = NULL;
    char *p;
    /* Flags for options */
    int ignore_root = 0;
    int nullok = 0;
    int not_set_pass = 0;
    pid_t code;
    int status, number = 0, pipedes[2];
    char *cargv[16];
    char *clog_program;


    clog_program = malloc (strlen(CLOG_PROGRAM)*sizeof(char));
    strcpy (clog_program,CLOG_PROGRAM);

/* Parse parameters for module */
    for (; argc-- > 0; argv++) {
	if (strcmp(*argv, "ignore_root") == 0)
	    ignore_root = 1;
        else if (strcmp(*argv,"clog") == 0) {
            if (argc-- > 0) {
            clog_program = realloc (clog_program,strlen(*argv++)*sizeof(char));
            strcpy (clog_program,*argv);
            }
	} else if (strcasecmp(*argv, "nullok") == 0)
	    nullok = 1;
	else if (strcasecmp(*argv, "not_set_pass") == 0)
	    not_set_pass = 1;
    }


    retval = pam_get_user(pamh, &user, NULL);
    if (retval != PAM_SUCCESS)
	return retval;

    if (user == NULL || !isalnum(*user)) {
	if (user)
	    syslog(LOG_ERR, "bad username [%s]", user);
	return PAM_USER_UNKNOWN;
    }

/* Get User Information */
    pw = getpwnam(user);

/* User does not exist */
    if (pw == NULL)
	return PAM_USER_UNKNOWN;

/* Handle Root */
    if (pw->pw_uid == 0 && ignore_root == 1)
	return (PAM_IGNORE);

    retval = pam_get_item(pamh, PAM_AUTHTOK, (void *) &p);
    if (retval != PAM_SUCCESS)
	return retval;

    if (p == NULL) {
	retval = __get_authtok(pamh, not_set_pass);
	if (retval != PAM_SUCCESS)
	    return retval;

	/* We have to call pam_get_item() again because value of p should
	   changed to a password. */

	pam_get_item(pamh, PAM_AUTHTOK, (void *) &p);
    }

/* If authtok is empty again, check if this is allowed */
    if (p == NULL) {
	if (flags & PAM_DISALLOW_NULL_AUTHTOK || !nullok)
	    return PAM_AUTH_ERR;
	return PAM_SUCCESS;
    }

/* set up pipe for communication with clog */
    if (pipe(pipedes) != 0) {
	syslog(LOG_CRIT, "pam_linux_afs: Can't open pipe");
	return (PAM_AUTHINFO_UNAVAIL);
    }

/* set up arguments for clog */
    cargv[number++] = "clog";
    cargv[number++] = "-pipe";
    cargv[number++] = user;
    cargv[number] = NULL;

    code = fork();
    if (code == -1) {
	syslog(LOG_ERR, "PAM-unix_auth: fork faild");
	return PAM_AUTH_ERR;
    }
    if (code == 0) {
	if (setuid(pw->pw_uid) < 0) {
	    syslog(LOG_ERR, "PAM-unix_auth: seteuid(%d) faild",
		   pw->pw_uid);
	    return (PAM_AUTHINFO_UNAVAIL);
	}
	close(0);		/* stdin */
	dup(pipedes[0]);
	close(pipedes[0]);
	close(1);		/* stdout */
	dup(pipedes[1]);
	close(pipedes[1]);
	close(2);
	execv(clog_program, cargv);
        memset(p, '\0', strlen(p));
	/* close pipe */
	close(0);		/* stdin */
	close(1);		/* stdout */
	/* Error: exec did return! */
	syslog(LOG_CRIT, "exec returned");
	return (PAM_AUTHINFO_UNAVAIL);
    }
/* write password */
    write(pipedes[1], p, strlen(p));
    write(pipedes[1], "\n", 1);
/* close pipe */
    close(pipedes[0]);
    close(pipedes[1]);
/* wait for child */
    if (code != wait(&status))
	return (PAM_AUTHINFO_UNAVAIL);
    user = NULL;		/* clean up */
    if (WIFEXITED(status))
	if (WEXITSTATUS(status) == 0)
	    return (PAM_SUCCESS);
    syslog(LOG_NOTICE, " pam_clog[%d]: unsuccessful",
	   getpid());
    return (PAM_AUTH_ERR);
}

PAM_EXTERN
int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc
                   , const char **argv)
{
   return (PAM_IGNORE);
}

/* end of module definition */

#ifdef PAM_STATIC

/* static module data */

struct pam_module _pam_coda_modstruct =
{
    "pam_coda",
    pam_sm_authenticate,
    pam_sm_setcred,
    NULL,
    NULL,
    NULL,
    NULL
};

#endif
